"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
// TODO make imports dynamic/only load what is needed for the method below
const Block = __importStar(require("multiformats/block"));
const cid_1 = require("multiformats/cid");
const dagPb = __importStar(require("@ipld/dag-pb"));
const dagCbor = __importStar(require("@ipld/dag-cbor"));
const raw = __importStar(require("multiformats/codecs/raw"));
const sha2_1 = require("multiformats/hashes/sha2");
const base58_1 = require("multiformats/bases/base58");
const peerbit_time_1 = require("@dao-xyz/peerbit-time");
const mhtype = "sha2-256";
const defaultBase = base58_1.base58btc;
const unsupportedCodecError = () => new Error("unsupported codec");
const BLOCK_TRANSPORT_TOPIC = "_block";
const cidifyString = (str) => {
    if (!str) {
        return str; // TODO fix types
    }
    return cid_1.CID.parse(str);
};
const stringifyCid = (cid) => {
    if (!cid || typeof cid === "string") {
        return cid;
    }
    const base = defaultBase;
    if (cid["/"]) {
        return cid["/"].toString(base);
    }
    return cid.toString(base);
};
const codecCodes = {
    [dagPb.code]: dagPb,
    [dagCbor.code]: dagCbor,
    [raw.code]: raw,
};
const codecMap = {
    raw: raw,
    "dag-pb": dagPb,
    "dag-cbor": dagCbor,
};
const _onMessage = (ipfs, me) => async (message) => {
    // TODO dont hardcode timeouts
    if (message.type === "signed") {
        if (message.from.toString() === me) {
            return;
        }
        const decoder = new TextDecoder();
        try {
            const cid = cidifyString(stringifyCid(decoder.decode(message.data)));
            const block = await ipfs.block.get(cid, {
                timeout: 1000,
            });
            await ipfs.pubsub.publish(BLOCK_TRANSPORT_TOPIC, block);
        }
        catch (error) {
            return; // timeout o r invalid cid
        }
    }
};
const announcePubSubBlocks = async (ipfs) => {
    const topics = await ipfs.pubsub.ls();
    if (topics.indexOf(BLOCK_TRANSPORT_TOPIC) === -1) {
        await ipfs.pubsub.subscribe(BLOCK_TRANSPORT_TOPIC, _onMessage(ipfs, (await ipfs.id()).id.toString()));
    }
};
const getBlockValue = (block, links) => {
    if (block.cid.code === dagPb.code) {
        return JSON.parse(new TextDecoder().decode(block.value.Data));
    }
    if (block.cid.code === dagCbor.code) {
        const value = block.value;
        links = links || [];
        links.forEach((prop) => {
            if (value[prop]) {
                value[prop] = Array.isArray(value[prop])
                    ? value[prop].map(stringifyCid)
                    : stringifyCid(value[prop]);
            }
        });
        return value;
    }
    if (block.cid.code === raw.code) {
        return block.value;
    }
    throw new Error("Unsupported");
};
async function readFromPubSub(ipfs, cid, options = {}) {
    const timeout = options.timeout || 5000;
    const cidString = stringifyCid(cid);
    const cidObject = cidifyString(cidString);
    const codec = codecCodes[cidObject.code];
    let value = undefined;
    const me = (await ipfs.id()).id.toString();
    const messageHandler = async (message) => {
        if (value) {
            return;
        }
        if (message.type === "signed") {
            if (message.from.toString() === me) {
                return;
            }
            const bytes = message.data;
            try {
                const decoded = await Block.decode({ bytes, codec, hasher: sha2_1.sha256 });
                if (!cidObject.equals(decoded.cid)) {
                    return;
                }
                value = getBlockValue(decoded, options.links);
            }
            catch (error) {
                // invalid bytes like "CBOR decode error: not enough data for type"
                return;
            }
        }
    };
    await ipfs.pubsub.subscribe(BLOCK_TRANSPORT_TOPIC, messageHandler, {
        timeout,
    });
    await ipfs.pubsub.publish(BLOCK_TRANSPORT_TOPIC, new TextEncoder().encode(cidString));
    try {
        await (0, peerbit_time_1.waitFor)(() => value !== undefined, {
            timeout,
            delayInterval: 100,
        });
    }
    catch (error) {
        /// TODO, timeout or?
    }
    finally {
        await ipfs.pubsub.unsubscribe(BLOCK_TRANSPORT_TOPIC, messageHandler);
    }
    return value;
}
async function readFromBlock(ipfs, cid, options = {}) {
    cid = cidifyString(stringifyCid(cid));
    const codec = codecCodes[cid.code];
    const bytes = await ipfs.block.get(cid, {
        timeout: options.timeout,
    });
    const block = await Block.decode({ bytes, codec, hasher: sha2_1.sha256 });
    return getBlockValue(block, options?.links);
}
async function read(ipfs, cid, options = {}) {
    const promises = [
        readFromPubSub(ipfs, cid, options),
        readFromBlock(ipfs, cid, options),
    ];
    const result = await Promise.any(promises);
    if (!result) {
        const results = await Promise.all(promises);
        for (const result of results) {
            if (result) {
                return result;
            }
        }
    }
    return result;
}
const prepareBlockWrite = (codec, value, links) => {
    if (!codec)
        throw unsupportedCodecError();
    if (codec.code === dagPb.code) {
        value = typeof value === "string" ? value : JSON.stringify(value);
        value = { Data: new TextEncoder().encode(value), Links: [] };
    }
    if (codec.code === dagCbor.code) {
        links = links || [];
        links.forEach((prop) => {
            if (value[prop]) {
                value[prop] = Array.isArray(value[prop])
                    ? value[prop].map(cidifyString)
                    : cidifyString(value[prop]);
            }
        });
    }
    return value;
};
async function write(ipfs, format, value, options = {}) {
    const codec = codecMap[format];
    value = prepareBlockWrite(codec, value, options?.links);
    const block = await Block.encode({ value, codec, hasher: sha2_1.sha256 });
    if (!options.onlyHash) {
        await ipfs.block.put(block.bytes, {
            /* cid: block.cid.bytes, */
            version: block.cid.version,
            format,
            mhtype,
            pin: options.pin,
            timeout: options.timeout,
        });
    }
    const cid = codec.code === dagPb.code ? block.cid.toV0() : block.cid;
    const cidString = cid.toString(options.base || defaultBase);
    await announcePubSubBlocks(ipfs);
    return cidString;
}
async function rm(ipfs, hash) {
    try {
        await ipfs.pin.rm(hash);
    }
    catch (error) {
        // not pinned // TODO add bettor error handling
    }
    for await (const result of ipfs.block.rm((hash instanceof cid_1.CID ? hash : cid_1.CID.parse(hash)))) {
        if (result.error) {
            throw new Error(`Failed to remove block ${result.cid} due to ${result.error.message}`);
        }
    }
}
exports.default = {
    read,
    write,
    readFromBlock,
    readFromPubSub,
    rm,
    BLOCK_TRANSPORT_TOPIC,
};
//# sourceMappingURL=index.js.map